%Direct Hamiltonian metho for solving an optimal control problem governed by the Fokker-Planck equaiton 
%implemented by Tim Breitenbach

function [] = DH()
OCP=struct('dx',0.1,'dt',0.002,'D',1/2,'a',-5,'b',5,'T',1,'umin',-10,'umax',10,'alpha',10^-5,'beta',10^-3);
dt=OCP.dt;
dx=OCP.dx;
alfax = -dt/dx;
alfay = -dt/dx;
N=(OCP.b-OCP.a)/dx;
Nt=round(OCP.T/dt);
tole=10^-12;
alpha=OCP.alpha;
beta=OCP.beta;

p=zeros(N+1,N+1);
pres=zeros(N+1,N+1,Nt);
u1=zeros(N-1,N-1,Nt+1); %the control is determined in each time step. Since \nabla p =0 in the final time (p=0), the Hamiltonian takes its minmum for u=0 at the final time 
u2=zeros(N-1,N-1,Nt+1);
upar1=zeros(N+1,N+1);
upar2=zeros(N+1,N+1);
f0=get_f0(0,OCP);
A=get_cd(OCP);

fl_cost_func=1; %Perform cost functional calculation at the beginning and the end of the DH method

if(fl_cost_func==1)
    u1_int=u1;
    u2_int=u2;
    u1=zeros(N+1,N+1,Nt+1);
    u2=zeros(N+1,N+1,Nt+1);
    u1(2:N,2:N,:)=u1_int;
    u2(2:N,2:N,:)=u2_int;
    f=fok_pl(u1,u2,f0,1,OCP);
    inti_J=get_J(u1,u2,f,A,OCP)
    u1=u1(2:N,2:N,:);
    u2=u2(2:N,2:N,:);
end
K=N+1;
map = inline('(k-1) * K + j','j','k','K');
tic
for i=Nt+1:-1:2
    %Assemble the matrix to calculate the adjoint variable for the next
    %time step
    i 
    upar1(2:N,2:N)=u1(:,:,i);
    upar2(2:N,2:N)=u2(:,:,i);
    for j=1:(N+1)
        for k=1:(N+1)
             B1(j,k) = -upar1(j,k);
             C1(j,k) = OCP.D;

             B2(j,k) = -upar2(j,k);
             C2(j,k) = OCP.D;
        
              w1 = dx*B1(j,k)/C1(j,k);
              if(w1>tole)
                  delta1(j,k)=1/w1 - 1/(exp(w1)-1);
              else
                  delta1(j,k)=0.5;
              end
              
              w2 = dx*B2(j,k)/C2(j,k);
              if(w2>tole)
                  delta2(j,k)=1/w2 - 1/(exp(w2)-1);
              else
                  delta2(j,k)=0.5;
              end
                     
        end
    end
    
    for j=1:(N+1)
        for k=1:(N+1)
            l = map(j,k,K);
        
        % coefficienti diagonali
        H(l,l) = 0;
        if ( j < N+1 ) 
            H(l,l) = H(l,l) + alfax*(C1(j,k)/dx - delta1(j,k)*B1(j,k));
        end
        if ( k < N+1 )
            H(l,l) = H(l,l) + alfay*(C2(j,k)/dx - delta2(j,k)*B2(j,k));  
        end
        if (j-1 >= 1 && j-1 < N+1 )
            H(l,l) = H(l,l) + alfax*(C1(j-1,k)/dx + ( 1 - delta1(j-1,k))*B1(j-1,k));
        end
        if (k-1 >= 1 && k-1 < N+1 )
            H(l,l) = H(l,l) + alfay*(C2(j,k-1)/dx + ( 1 - delta2(j,k-1))*B2(j,k-1));
        end        

        % coefficienti non-diagonali
        l1 = map(j+1,k,K);
        if ( j+1 <= N+1 )
            if ( j < N+1 )    
                H(l,l1) = -alfax*((1-delta1(j,k))*B1(j,k) + C1(j,k)/dx);
            end
        end
        
        l1 = map(j,k+1,K);
        if ( k+1 <= N+1 )
            if ( k < N+1 )    
                H(l,l1) = -alfay*((1-delta2(j,k))*B2(j,k) + C2(j,k)/dx);
            end
        end       

        l1 = map(j-1,k,K);
        if ( j-1 >= 1 )
                H(l,l1) = -alfax*(-delta1(j-1,k)*B1(j-1,k) + C1(j-1,k)/dx);
        end
        
        l1 = map(j,k-1,K);
        if ( k-1 >= 1 )
                H(l,l1) = -alfay*(-delta2(j,k-1)*B2(j,k-1) + C2(j,k-1)/dx);
        end       
        end
    end
    
    H=speye((N+1)*(N+1),(N+1)*(N+1))+H;
    H=H';
 
    
   
    ph=mat_to_vec(p)';
    ph=H*ph; %Calculate the adjoint varialbe for the next time step (backwards)
  
    ph=vect_to_mtx((N+1)*(N+1),N+1,ph);
    ph(2:N,2:N)=ph(2:N,2:N)+dt*A(2:N,2:N,i-1)+dt*0.5*(OCP.alpha)*(u1(:,:,i).^2+u2(:,:,i).^2)+dt*OCP.beta*(abs(u1(:,:,i))+abs(u2(:,:,i)));
      if(sum(sum(abs(ph)))*dx*dx>10^3)
          fprintf('int p too big: possibly instability\n')
      end
    p=ph;
    
    
    pres(:,:,i-1)=p;
    %Approximate the space derivative of the adjoint variable in the
    %Hamiltonian
    dp1=(p(3:N+1,2:N)-p(2:N,2:N))/(dx);
    dp2=(p(2:N,3:N+1)-p(2:N,2:N))/(dx);
            
     %Calculate the control for the next time step.
    
    u11=min(max(0,(-dp1-OCP.beta)/OCP.alpha),OCP.umax);
    u12=min(max(OCP.umin,(-dp1+OCP.beta)/OCP.alpha),0);
    u21=min(max(0,(-dp2-OCP.beta)/OCP.alpha),OCP.umax);
    u22=min(max(OCP.umin,(-dp2+OCP.beta)/OCP.alpha),0);
    
    for j=1:N-1
        for k=1:N-1
             VecVal=combvec([u11(j,k) u12(j,k)],[u21(j,k) u22(j,k)])';

            
            H1=0.5*alpha*(VecVal(:,1).^2 + VecVal(:,2).^2) + beta*(abs(VecVal(:,1))+abs(VecVal(:,2))) ...
                    +VecVal(:,1)*dp1(j,k)+VecVal(:,2)*dp2(j,k);
            [Hh(j,k),pos]=min(H1);
            u1h(j,k)=VecVal(pos,1);
            u2h(j,k)=VecVal(pos,2);
            
        end
    end
    Hres(:,:,i-1)=Hh;
    u1(:,:,i-1)=u1h;
    u2(:,:,i-1)=u2h;
            
end
toc
u1h=u1;
u2h=u2;
u1=zeros(N+1,N+1,Nt+1);
u2=zeros(N+1,N+1,Nt+1);
u1(2:N,2:N,:)=u1h;
u2(2:N,2:N,:)=u2h;

if(fl_cost_func==1)
     f=fok_pl(u1,u2,f0,1,OCP);
     final_J=get_J(u1,u2,f,A,OCP)
end




[a,b,m,h,x,y,Nt,t,dt,alpha,beta,nu] = parameters(1,OCP);
Nx = m+1;
ip = traj1(t(1)); 
T = t(Nt);
T1 = T;
T0 = t(1);
fname = 'DH_CF.mat';
save(fname,'a','b','Nx','T0','T1','Nt','u1','u2','ip');


end

function uu = vect_to_mtx(NN,Nx,u)
    % ripristina u in formato matrice
    for l=0:NN-1
        j=mod(l,Nx);
        k=(l-j)/Nx;
        uu(j+1,k+1) = u(l+1); % verificare qui !!!!
    end
   uu=uu';
end
