%SQH method for solving an optimal control problem gouverned by the FP
%equation implemented by Tim Breitenbach

function [ ] = SQH_FP_u( )
OCP=struct('dx',0.1,'dt',0.002,'D',1/2,'a',-5,'b',5,'T',1,'umin',-10,'umax',10,'alpha',10^-5,'beta',10^-3);
 f0=get_f0(0,OCP);
% %surf(f0)

cdes=get_cd(OCP);
% %showcd(cdes,OCP);

Nt=OCP.T/OCP.dt;
N=((OCP.b-OCP.a)/OCP.dx);

u_1=0*ones(N+1,N+1,Nt);
u_2=0*ones(N+1,N+1,Nt);
u_1_old=u_1;
u_2_old=u_2;

f=fok_pl(u_1,u_2,f0,1,OCP);
f_old=f;
 %showf(f,OCP);


%sum(sum(f(:,:,end)))*(OCP.dx^2)

p=adj_fok_pl(u_1,u_2,cdes,1,OCP);
% showp(p,OCP)

J_plot(1,1)=get_J(u_1,u_2,f,cdes,OCP);
Jstart=J_plot(1,1)
Jk(1)=J_plot(1,1);
kmax=20000;



dx=OCP.dx;
epsi=10^-2;
count_updates=1;

tic
for k=1:kmax
for i=1:Nt
    if(i==1)
        fint=f0(2:N,2:N);
    else
        fint=f(2:N,2:N,i);
    end
    %Aproximation of the derivative of the adjoint variable in the
    %augmented Hamiltonian
    dp1=(p(3:N+1,2:N,i)-p(2:N,2:N,i))/(dx);
    dp2=(p(2:N,3:N+1,i)-p(2:N,2:N,i))/(dx);
    
    u11=min(max(0,(2*epsi*u_1(2:N,2:N,i)-dp1.*fint-OCP.beta.*fint)./(2*epsi+OCP.alpha.*fint)),OCP.umax);
    u12=min(max(OCP.umin,(2*epsi*u_1(2:N,2:N,i)-dp1.*fint+OCP.beta.*fint)./(2*epsi+OCP.alpha.*fint)),0);
    u21=min(max(0,(2*epsi*u_2(2:N,2:N,i)-dp2.*fint-OCP.beta.*fint)./(2*epsi+OCP.alpha.*fint)),OCP.umax);
    u22=min(max(OCP.umin,(2*epsi*u_2(2:N,2:N,i)-dp2.*fint+OCP.beta.*fint)./(2*epsi+OCP.alpha.*fint)),0);
    for j=1:N-1
        for l=1:N-1
            VecVal=combvec([u11(j,l) u12(j,l)],[u21(j,l) u22(j,l)])';
            H1=(0.5*OCP.alpha*(VecVal(:,1).^2 + VecVal(:,2).^2) + OCP.beta*(abs(VecVal(:,1))+abs(VecVal(:,2))) ...
                    +VecVal(:,1)*dp1(j,l)+VecVal(:,2)*dp2(j,l))*fint(j,l)...
                      +epsi*((VecVal(:,1)-u_1(j+1,l+1,i)).^2 + (VecVal(:,2)-u_2(j+1,l+1,i)).^2);
                    
       
            [~,pos]=min(H1);
            u_1(j+1,l+1,i)=VecVal(pos,1);
            u_2(j+1,l+1,i)=VecVal(pos,2);
        end
    end
    
        
end
    
    
   intu=sum(sum(sum(((u_1-u_1_old).^2)+((u_2-u_2_old).^2))))*OCP.dt*(OCP.dx^2)
   
   f=fok_pl(u_1,u_2,f0,1,OCP);
   J_int=get_J(u_1,u_2,f,cdes,OCP);
   if(J_int-Jk(count_updates)>-(10^-9)*intu)
       u_1=u_1_old;
       u_2=u_2_old;
       f=f_old;
       epsi=epsi*50                            
   else
       count_updates=count_updates+1;
       p=adj_fok_pl(u_1,u_2,cdes,1,OCP);
       epsi=epsi*0.15                       
       J_int
       Jk(count_updates)=J_int;
       u_1_old=u_1;
       u_2_old=u_2;
       f_old=f;
       
       u1=u_1;
       u2=u_2;  
       
   end
   [a,b,m,h,x,y,Nt,t,dt,alpha,beta,nu] = parameters(1,OCP);
    Nx = m+1;
    ip = traj1(t(1)); 
    T = t(Nt);
    T1 = T;
    T0 = t(1);
    fname = 'SQH_CF.mat';
    save(fname,'a','b','Nx','T0','T1','Nt','u1','u2','ip');
   
   if(intu<10^-10)
        k
        u_1=u_1_old;
        u_2=u_2_old;
        break;
   end 
   
   [a,b,m,h,x,y,Nt,t,dt,alpha,beta,nu] = parameters(1,OCP);
    Nx = m+1;
    ip = traj1(t(1)); 
    T = t(Nt);
    T1 = T;
    T0 = t(1);
    fname = 'SQH_CF.mat';
    save(fname,'a','b','Nx','T0','T1','Nt','u1','u2','ip');
   
end 
toc

end

