%Implementation of the SQH method for a time dependend control mechanism by
%Tim Breitenbach

function [ ] = SQH_FP( )
OCP=struct('dx',0.1,'dt',0.01,'D',1/32,'a',-5,'b',5,'T',1,'vmin',-2,'vmax',2,'wmin',-1,'wmax',1,'sv',0.6,'sw',0.3,'alpha',0.5*10^(-2)+0*15/10000, 'beta',0.5*10^(-2)+0*15/10000);
f0=get_f0(0,OCP);
%surf(f0)

cdes=get_cd(OCP);
%showcd(cdes,OCP);

Nt=OCP.T/OCP.dt;
N=((OCP.b-OCP.a)/OCP.dx)-1;

v=0*2*ones(2,Nt);
w=0*1*ones(2,Nt);
v_old=v;
w_old=w;

%PMP test for the initial guess of the controls
optimal_exact(v,w,f0,cdes,OCP)
optimal(v,w,f0,cdes,OCP)

 f=fok_pl(v,w,f0,1,OCP);
 f_old=f;
 %showf(f,OCP);


sum(sum(f(:,:,end)))*(OCP.dx^2)

p=adj_fok_pl(v,w,cdes,1,OCP);
%showp(p,OCP);



J_plot(1,1)=get_J(v,w,f,cdes,OCP);
J_plot(1,1)
Jk(1)=J_plot(1,1);
kmax=2000;



dx=OCP.dx;
sv=OCP.sv;
sw=OCP.sw;
epsi=10^-2;
count_updates=1;
Vs=(OCP.b-OCP.a)^2;

for k=1:kmax
    for i=1:Nt
       fzw=f(:,:,i);
       a=sum(sum(fzw))*dx*dx;
       dp_x1=(p(3:N+2,:,i)-p(1:N,:,i))/(2*dx);
       dp_x2=(p(:,3:N+2,i)-p(:,1:N,i))/(2*dx);
       b1=sum(sum(dp_x1.*fzw(2:N+1,:)))*dx*dx;
       b2=sum(sum(dp_x2.*fzw(:,2:N+1)))*dx*dx;
       c1=sum(sum((dp_x1.*fzw(2:N+1,:)).*((ones(N+2,1)*(OCP.a+(1:N)*dx)))'))*dx*dx;
       c2=sum(sum((dp_x2.*fzw(:,2:N+1)).*((ones(N+2,1)*(OCP.a+(1:N)*dx)))))*dx*dx;
       
       %Possible values where the augmented Hamiltoninan takes its minimum
       v11=min(max(sv,(2*epsi*Vs*v(1,i)-a*OCP.alpha-b1)/(2*Vs*epsi)),OCP.vmax);
       v12=min(max(-sv,(2*epsi*Vs*v(1,i)-b1)/(2*Vs*epsi)),sv);
       v13=min(max(OCP.vmin,(2*epsi*Vs*v(1,i)+a*OCP.alpha-b1)/(2*Vs*epsi)),-sv);
       
       v21=min(max(sv,(2*epsi*Vs*v(2,i)-a*OCP.alpha-b2)/(2*Vs*epsi)),OCP.vmax);
       v22=min(max(-sv,(2*epsi*Vs*v(2,i)-b2)/(2*Vs*epsi)),sv);
       v23=min(max(OCP.vmin,(2*epsi*Vs*v(2,i)+a*OCP.alpha-b2)/(2*Vs*epsi)),-sv);
       
       w11=min(max(sw,(2*epsi*Vs*w(1,i)-c1-a*OCP.beta)/(2*Vs*epsi)),OCP.wmax);
       w12=min(max(-sw,(2*epsi*Vs*w(1,i)-c1)/(2*Vs*epsi)),sw);
       w13=min(max(OCP.wmin,(2*epsi*Vs*w(1,i)-c1+a*OCP.beta)/(2*Vs*epsi)),-sw);
       
       w21=min(max(sw,(2*epsi*Vs*w(2,i)-c2-a*OCP.beta)/(2*Vs*epsi)),OCP.wmax);
       w22=min(max(-sw,(2*epsi*Vs*w(2,i)-c2)/(2*Vs*epsi)),sw);
       w23=min(max(OCP.wmin,(2*epsi*Vs*w(2,i)-c2+a*OCP.beta)/(2*Vs*epsi)),-sw);
       
       VecVal=combvec([v11 v12 v13],[v21 v22 v23], [w11 w12 w13], [w21 w22 w23])';
       

       H1=a*(OCP.alpha*(max(abs(VecVal(:,1))-OCP.sv,0)+max(abs(VecVal(:,2))-OCP.sv,0)) + ...
            OCP.beta*(max(abs(VecVal(:,3))-OCP.sw,0)+max(abs(VecVal(:,4))-OCP.sw,0)))...
           +b1*VecVal(:,1)+c1*VecVal(:,3)+b2*VecVal(:,2)+c2*VecVal(:,4)...
           +((OCP.b-OCP.a)^2)*epsi*( ((VecVal(:,1)-v(1,i)).^2) + ((VecVal(:,3)-w(1,i)).^2) + ((VecVal(:,2)-v(2,i)).^2) + (VecVal(:,4)-w(2,i)).^2);
       
        [~,pos]=min(H1);
        v(1,i)=VecVal(pos,1);
        v(2,i)=VecVal(pos,2);
        w(1,i)=VecVal(pos,3);
        w(2,i)=VecVal(pos,4);
   
    end
    
   intvw=sum((((v(1,:)-v_old(1,:)).^2)+((v(2,:)-v_old(2,:)).^2)+(w(1,:)-w_old(1,:)).^2)+((w(2,:)-w_old(2,:)).^2))*OCP.dt%*(OCP.b-OCP.a)^2
   
   f=fok_pl(v,w,f0,1,OCP);
   J_int=get_J(v,w,f,cdes,OCP);
   if(J_int-Jk(count_updates)>-(10^-7)*intvw)
       v=v_old;
       w=w_old;
       f=f_old;
       epsi=epsi*50;                           
   else
       count_updates=count_updates+1;
       p=adj_fok_pl(v,w,cdes,1,OCP);
       epsi=epsi*0.15;                        
       Jk(count_updates)=J_int;
       p_value(count_updates)=sum(sum(p(:,:,1).*f0))*OCP.dx^2;
       v_old=v;
       w_old=w;
       f_old=f;
   end
   if(intvw<10^-10)
        k
        v=v_old;
        w=w_old;
        break;
   end 
end

%Comparision of the Value function and the cost functional at the initial
%time
fprintf('int(p(0)f0),\t J\n\n')
for i=1:count_updates
    fprintf('%d \t %d\n',p_value(i),Jk(i))
end

%PMP test for the result of the SQH method
optimal_exact(v,w,f0,cdes,OCP)
optimal(v,w,f0,cdes,OCP)



figure
subplot(2,2,1)
plot(0:OCP.dt:OCP.T-OCP.dt,v(1,:))
title('$v_1$','interpreter','latex');
subplot(2,2,2)
plot(0:OCP.dt:OCP.T-OCP.dt,v(2,:))
title('$v_2$','interpreter','latex')
subplot(2,2,3)
plot(0:OCP.dt:OCP.T-OCP.dt,w(1,:))
title('$w_1$','interpreter','latex')
subplot(2,2,4)
plot(0:OCP.dt:OCP.T-OCP.dt,w(2,:))
title('$w_2$','interpreter','latex')

show(cdes,f,p,OCP)

sum(sum(f(:,:,1)))*OCP.dx*OCP.dx
sum(sum(f(:,:,end)))*OCP.dx*OCP.dx

figure
plot(Jk)



T=OCP.T;
Nt=T/OCP.dt;
N=((OCP.b-OCP.a)/OCP.dx);
x1 = linspace(OCP.a,OCP.b,N+1);
x2 = linspace(OCP.a,OCP.b,N+1);
u_1=zeros(N+1,N+1,Nt);
u_2=zeros(N+1,N+1,Nt);
%Construction of the control from v and w
for i=1:Nt
    u1(:,:,i)=(ones(N+1,N+1)*v(1,i)+((ones(N+1,1)*x1)')*w(1,i));
    u2(:,:,i)=(ones(N+1,N+1)*v(2,i)+((ones(N+1,1)*x2))*w(2,i));
end

%Calculation of the mean value of f
ffp(:,:,1)=f0;
ffp(:,:,2:Nt+1)=f;
for i=1:Nt+1
    fzwi=ffp(:,:,i);
    zwsumme1=0;
    zwsumme2=0;
    for j=1:N+1
         zwsumme1=zwsumme1+(-5+dx*(j-1))*sum(fzwi(j,:))*dx^2;
         zwsumme2=zwsumme2+(-5+dx*(j-1))*sum(fzwi(:,j))*dx^2;
    end
    mv(1,i)=zwsumme1;
    mv(2,i)=zwsumme2;
end
figure 
hold on
plot(mv(1,:),mv(2,:),'--mo');

%Save the control values
[a,b,m,h,x,y,Nt,t,dt,alpha,beta,nu] = parameters(1,OCP);
Nx = m+1;
ip = traj1(t(1)); 
T = t(Nt);
T1 = T;
T0 = t(1);
fname = 'data_traj_split.mat';
save(fname,'a','b','Nx','T0','T1','Nt','u1','u2','ip');

end

