%Implementation of the SQH method for an optimal control problem with  stepwise costs for the control by Tim Breitenbach
function [ ] =SQH_stepCost(  )
%Problem parameters
OCP=struct('N',200,'Nt',200,'a',0,'b',1,'T',1,'D',0.2,'u_lo',-10,'u_up',10,'d_KU',1/100,'alpha',0,'beta',10^-1,'P',0,'s',3);

[X,T]=meshgrid(OCP.a:(OCP.b-OCP.a)/OCP.N:OCP.b,0:OCP.T/OCP.Nt:OCP.T);
[Xu,Tu]=meshgrid((OCP.a+(OCP.b-OCP.a)/OCP.N):(OCP.b-OCP.a)/OCP.N:(OCP.b-(OCP.b-OCP.a)/OCP.N),0+OCP.T/OCP.Nt:OCP.T/OCP.Nt:OCP.T);

f=zeros(OCP.Nt*(OCP.N+1),1); %f(x,t) further right hand-side of the state equation

hx=(OCP.b-OCP.a)/OCP.N; 
dt=OCP.T/OCP.Nt; 

y0=zeros((OCP.N+1),1);

yd=desiredState2(OCP);

 figure('Name','yd')
 intyd=reshape(yd,[OCP.N+1,OCP.Nt+1])';
 mesh(X,T,intyd)
  xlabel('x');
  ylabel('t');
 view(-55,15);


u=0*ones((OCP.N-1)*OCP.Nt,1); %Initial guess for the control
u_old=u;
y=implForward(u,y0,f,OCP);
y_old=y;
p=implBackward(y,yd,OCP);

J_plot(1,1)=get_J(y,yd,u,OCP );
Jk(1)=J_plot(1,1);
kmax=2000; %Maximum iteration number

epsi=0.6*10^-0;
count_updates=1;
tic
for k=1:kmax
    for j=1:OCP.Nt
        for i=1:OCP.N-1
            %Minimizing the augmented Hamiltonian pointwise 
            u((j-1)*(OCP.N-1)+i,1)=argminH(i,j,p,u((j-1)*(OCP.N-1)+i,1),epsi,f,OCP);   

        end
    end   
    
   du=sum((u-u_old).^2)*hx*dt; %Norm of the control update
   
   y=implForward(u,y0,f,OCP);
   J_int=get_J(y,yd,u,OCP );
   if(J_int-Jk(count_updates)>-(10^-9)*du)
       u=u_old;
       y=y_old;
       epsi=epsi*50;                           
   else
       count_updates=count_updates+1;
       p=implBackward(y,yd,OCP);
       epsi=epsi*0.15;                     
       Jk(count_updates)=J_int;
       u_old=u;
       y_old=y;
      
   end
   if(du<10^-6)
        k
        u=u_old;
        break;
   end 
   
end
toc
y=implForward(u,y0,f,OCP);

 figure('Name','y')
 inty=reshape(y,[OCP.N+1,OCP.Nt+1])';
 mesh(X,T,inty)
 xlabel('x');
 ylabel('t');
 view(-55,15);

 figure('Name','u')
 intu=reshape(u,[OCP.N-1,OCP.Nt])';
 mesh(Xu,Tu,intu)
 xlabel('x');
 ylabel('t');
 view(-55,15);
 
 figure
 contour(Xu,Tu,intu,20)
 xlabel('x');
 ylabel('t');
 
 figure
 plot(Jk)
 
check_u_opt( u,y0,f,yd,OCP ); %PMP optimality test
end

