%
%The function calculates the solution to the adjoint equation of a linear parabolic optimization problem
%-p'-D\Delta p =y-yd on (a,b)x(0,T) with an implicit Euler scheme, see corresponding paper Subsection 4.1 for details

function [ p] = backward(y,yd, OCP)

N=OCP.N;            %Number of intervals in space
h=(OCP.b-OCP.a)/N;
Nt=OCP.Nt;          %Number of intervals in time 
dt=OCP.T/Nt;
D=OCP.D;            %Diffusion constant

p=zeros((N+1)*(Nt+1),1);
%Assemble one dimensional discretised Laplacian
e=ones(N-1,1);      
ed=(D*dt/h^2)*e;
A=spdiags([-ed e+2*ed -ed],-1:1, N-1,N-1);

for j=Nt:-1:1
    %The update for each time step according to an implicit Euler scheme
    p((j-1)*(N+1)+2:j*(N+1)-1,1)=A\(p(j*(N+1)+2:(j+1)*(N+1)-1,1)...
                                +dt*(y((j)*(N+1)+2:(j+1)*(N+1)-1)-yd((j)*(N+1)+2:(j+1)*(N+1)-1))); 
end

end

