%
%The function uses an implicit Euler scheme to solve a linear heat equation with right hand side
%y'-D\Delta y = u+f, see corresponding paper Subsection 4.1 for details

function [ y ] = forward_y(u,y0,f,OCP)

Nt=OCP.Nt;          %Number of intervals in time 
dt=OCP.T/Nt;
N=OCP.N;            %Number of intervals in space
h=(OCP.b-OCP.a)/N;
D=OCP.D;            %Diffusion constant

y=[y0;zeros((N+1)*Nt,1)];
%Assemble one dimensional discretised Laplacian
e=ones(N-1,1); 
ed=(D*dt/h^2)*e;
A=spdiags([-ed e+2*ed -ed],-1:1, N-1,N-1);

for j=2:Nt+1
    %The update for each time step according to an implicit Euler scheme
    y((j-1)*(N+1)+2:j*(N+1)-1,1)=A\(y((j-2)*(N+1)+2:(j-1)*(N+1)-1,1)+dt*(u((j-2)*(N-1)+1:(j-1)*(N-1),1)...
                                +f((j-2)*(N+1)+2:(j-1)*(N+1)-1,1))); 
end

end

