%
%The function solves the state equation of an elliptic linear equation with zero boundary conditions 
%and with max(y,0) in the state equation, see corresponding paper, Subsection 4.4 for details


function [ yit ] = forward_y(f,OCP)

%Assemble the Laplacian
h=(OCP.b-OCP.a)/(OCP.N+1); 
n=OCP.N+1;
N=(n+1)*(n+1)-4*n;
v=ones(N,1);
v_block=[ones(n-2,1);0];
v_block=kron(ones(N/(n-1),1),v_block);
A_hilf=spdiags([(-1/h^2)*v_block (-1/h^2)*[v_block(N); v_block(1:N-1)]],[ -1  1 ],N,N);
A=spdiags([(-1/h^2)*v  (4/h^2)*v  (-1/h^2)*v],[-(n-1)  0  n-1],N,N);
A=A+A_hilf;

 yit=zeros(OCP.N+2,OCP.N+2);
 for k=1:100000
 %Loop over all interior grid points
 for i=2:OCP.N+1    
     for j=2:OCP.N+1
         %Assemble right hand side
         rhs_disc=((yit(i+1,j)+yit(i-1,j)+yit(i,j+1)+yit(i,j-1))/h^2)+f(i-1,j-1); 
         if(rhs_disc>=0)  
             %Update the state variable dependend on the sign of the right hand side
             yit(i,j)=rhs_disc/((4/h^2)+1);
         else
             yit(i,j)=rhs_disc/((4/h^2));
         end
     end
 end
 %Determine the residuum of the state equation
 res=norm(A*reshape(yit(2:OCP.N+1,2:OCP.N+1)',[OCP.N*OCP.N,1])+max(0,reshape(yit(2:OCP.N+1,2:OCP.N+1)',[OCP.N*OCP.N,1]))...
     -reshape(f',[OCP.N*OCP.N,1]))*h; 
     if(res<10^-6)  
         %Stop if the residuum of the state equation is less than 10^-8
         break;
     end
 end
 
end

