%Numerical PMP test where the minimum of the Hamiltonian is determined with
%the formulas with which the augmented Hamiltonian is minimized in the main
%file

function [] = optimal(v,w,fin,cd,OCP)

f=fok_pl(v,w,fin,1,OCP);
p=adj_fok_pl(v,w,cd,1,OCP);
Nt=OCP.T/OCP.dt;
dx=OCP.dx;
N=((OCP.b-OCP.a)/OCP.dx)-1;
sv=OCP.sv;
sw=OCP.sw;
epsi=0;


for i=1:Nt
       fzw=f(:,:,i);
       a=sum(sum(fzw))*dx*dx;
       dp_x1=(p(3:N+2,:,i)-p(1:N,:,i))/(2*dx);
       dp_x2=(p(:,3:N+2,i)-p(:,1:N,i))/(2*dx);
       b1=sum(sum(dp_x1.*fzw(2:N+1,:)))*dx*dx;
       b2=sum(sum(dp_x2.*fzw(:,2:N+1)))*dx*dx;
       c1=sum(sum((dp_x1.*fzw(2:N+1,:)).*((ones(N+2,1)*(OCP.a+(1:N)*dx)))'))*dx*dx;
       c2=sum(sum((dp_x2.*fzw(:,2:N+1)).*((ones(N+2,1)*(OCP.a+(1:N)*dx)))))*dx*dx;
       Hopt(i)=a*(OCP.alpha*(max(abs(v(1,i))-OCP.sv,0)+max(abs(v(2,i))-OCP.sv,0)) + ...
            OCP.beta*(max(abs(w(1,i))-OCP.sw,0)+max(abs(w(2,i))-OCP.sw,0)))...
           +b1*v(1,i)+c1*w(1,i)+b2*v(2,i)+c2*w(2,i);
           
end

for i=1:Nt
       fzw=f(:,:,i);
       a=sum(sum(fzw))*dx*dx;
       dp_x1=(p(3:N+2,:,i)-p(1:N,:,i))/(2*dx);
       dp_x2=(p(:,3:N+2,i)-p(:,1:N,i))/(2*dx);
       b1=sum(sum(dp_x1.*fzw(2:N+1,:)))*dx*dx;
       b2=sum(sum(dp_x2.*fzw(:,2:N+1)))*dx*dx;
       c1=sum(sum((dp_x1.*fzw(2:N+1,:)).*((ones(N+2,1)*(OCP.a+(1:N)*dx)))'))*dx*dx;
       c2=sum(sum((dp_x2.*fzw(:,2:N+1)).*((ones(N+2,1)*(OCP.a+(1:N)*dx)))))*dx*dx;
       
       v11=min(max(sv,(2*epsi*v(1,i)-a*OCP.alpha-b1)/(2*epsi)),OCP.vmax);
       v12=min(max(-sv,(2*epsi*v(1,i)-b1)/(2*epsi)),sv);
       v13=min(max(OCP.vmin,(2*epsi*v(1,i)+a*OCP.alpha-b1)/(2*epsi)),-sv);
       
       v21=min(max(sv,(2*epsi*v(2,i)-a*OCP.alpha-b2)/(2*epsi)),OCP.vmax);
       v22=min(max(-sv,(2*epsi*v(2,i)-b2)/(2*epsi)),sv);
       v23=min(max(OCP.vmin,(2*epsi*v(2,i)+a*OCP.alpha-b2)/(2*epsi)),-sv);
       
       w11=min(max(sw,(2*epsi*w(1,i)-c1-a*OCP.beta)/(2*epsi)),OCP.wmax);
       w12=min(max(-sw,(2*epsi*w(1,i)-c1)/(2*epsi)),sw);
       w13=min(max(OCP.wmin,(2*epsi*w(1,i)-c1+a*OCP.beta)/(2*epsi)),-sw);
       
       w21=min(max(sw,(2*epsi*w(2,i)-c2-a*OCP.beta)/(2*epsi)),OCP.wmax);
       w22=min(max(-sw,(2*epsi*w(2,i)-c2)/(2*epsi)),sw);
       w23=min(max(OCP.wmin,(2*epsi*w(2,i)-c2+a*OCP.beta)/(2*epsi)),-sw);
       
       VecVal=combvec([v11 v12 v13],[v21 v22 v23], [w11 w12 w13], [w21 w22 w23])';
       

       H1=a*(OCP.alpha*(max(abs(VecVal(:,1))-OCP.sv,0)+max(abs(VecVal(:,2))-OCP.sv,0)) + ...
            OCP.beta*(max(abs(VecVal(:,3))-OCP.sw,0)+max(abs(VecVal(:,4))-OCP.sw,0)))...
           +b1*VecVal(:,1)+c1*VecVal(:,3)+b2*VecVal(:,2)+c2*VecVal(:,4)...
           +epsi*( ((VecVal(:,1)-v(1,i)).^2) + ((VecVal(:,3)-w(1,i)).^2) + ((VecVal(:,2)-v(2,i)).^2) + (VecVal(:,4)-w(2,i)).^2);
       
        [Hnew(i),pos]=min(H1);
        v(1,i)=VecVal(pos,1);
        v(2,i)=VecVal(pos,2);
        w(1,i)=VecVal(pos,3);
        w(2,i)=VecVal(pos,4);
end

fprintf('Max DH: %d\n',max(Hopt-Hnew));
fprintf('Min DH: %d\n',min(Hopt-Hnew));
fprintf('Fraction of grid points where 0<=DH<=10^-l is fulfilled:\n')
for l=1:12
count=0;
for i=1:Nt
         %Count the numer of grid points where the solution u is PMP optimal up to the tolerance 10^-l
            if (Hopt(i)-Hnew(i)<=10^-l)
                count=count+1;
            end
end

ratio=count/Nt;
fprintf('l=%i: %f\t',l,ratio);
end
fprintf('\n');

end

